"""Provide utilities to be used by interfaces."""
import asyncio
import shlex
from typing import List

from lm_agent.config import settings
from lm_agent.exceptions import CommandFailedToExecute
from lm_agent.logs import logger


async def run_command(command_line_parts: List[str]) -> str:
    """
    Run a command using a subprocess shell.
    Returns the output as string if the command succeeds.
    Raises CommandFailedToExecute exception if return code is not zero.
    """

    command_line = shlex.join(command_line_parts)
    proc = await asyncio.create_subprocess_shell(
        command_line, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.STDOUT
    )

    # block until the command succeeds
    stdout, _ = await asyncio.wait_for(proc.communicate(), settings.TOOL_TIMEOUT)
    output = str(stdout, encoding=settings.ENCODING, errors="replace")

    if proc.returncode != 0:
        error_message = shlex.join(
            [
                f"Command {command_line} failed!",
                f"Error: {output}",
                f"Return code: {proc.returncode}",
            ]
        )
        logger.error(error_message)
        raise CommandFailedToExecute(f"The command failed to execute, with return code {proc.returncode}.")

    return output



#!/usr/bin/env python3
"""
The EpilogSlurmctld executable.
"""
import asyncio
import sys

from lm_agent.backend_utils.utils import remove_job_by_slurm_job_id
from lm_agent.config import settings
from lm_agent.logs import init_logging, logger
from lm_agent.services.reconciliation import reconcile
from lm_agent.workload_managers.slurm.cmd_utils import get_required_licenses_for_job
from lm_agent.workload_managers.slurm.common import get_job_context


async def epilog():
    # Initialize the logger
    init_logging("slurmctld-epilog")
    job_context = await get_job_context()
    job_id = job_context["job_id"]
    job_licenses = job_context["job_licenses"]

    # Check if reconciliation should be triggered.
    if settings.USE_RECONCILE_IN_PROLOG_EPILOG:
        # Force a reconciliation before we attempt to remove bookings.
        try:
            await reconcile()
        except Exception as e:
            logger.error(f"Failed to call reconcile with {e}")
            sys.exit(1)

    try:
        required_licenses = get_required_licenses_for_job(job_licenses)
    except Exception as e:
        logger.error(f"Failed to call get_required_licenses_for_job with {e}")
        sys.exit(1)

    if not required_licenses:
        logger.debug(f"No licenses required for job {job_id}, exiting!")
        sys.exit(0)

    if len(required_licenses) > 0:
        # Attempt to remove the job with its bookings.
        await remove_job_by_slurm_job_id(job_id)
        logger.debug(f"Job {job_id} removed successfully")


def main():
    asyncio.run(epilog())


if __name__ == "__main__":
    main()



"""Slurm reservation CRUD module."""
from typing import Union

from lm_agent.config import settings
from lm_agent.exceptions import CommandFailedToExecute, LicenseManagerReservationFailure
from lm_agent.logs import logger
from lm_agent.utils import run_command


async def scontrol_create_reservation(licenses: str, duration: str) -> bool:
    """
    Create the reservation for licenses with the specified duration.

    Duration format: [days-]hours:minutes:seconds
    Licenses format: foo.bar@server:123,foo.baz@server:234

    Returns True if the reservation was created successfully, otherwise returns False.
    """
    cmd = [
        str(settings.SCONTROL_PATH),
        "create",
        "reservation",
        f"user={settings.LM_USER}",
        f"reservation={settings.RESERVATION_IDENTIFIER}",
        "starttime=now",
        f"duration={duration}",
        "flags=license_only",
        f"licenses={licenses}",
    ]

    logger.debug(f"#### Creating reservation for {licenses} with duration {duration} ####")
    try:
        await run_command(cmd)
    except CommandFailedToExecute:
        logger.error(f"#### Failed to create reservation {settings.RESERVATION_IDENTIFIER} ####")
        return False

    logger.debug(f"#### Successfully created reservation {settings.RESERVATION_IDENTIFIER} ####")
    return True


async def scontrol_show_reservation() -> Union[str, bool]:
    """
    Read reservation from the cluster.

    Returns the reservation information if it exists, otherwise returns None.
    """
    cmd = [
        str(settings.SCONTROL_PATH),
        "show",
        "reservation",
        settings.RESERVATION_IDENTIFIER,
    ]

    logger.debug(f"#### Getting reservation {settings.RESERVATION_IDENTIFIER} ####")
    try:
        reservation_output = await run_command(cmd)
    except CommandFailedToExecute:
        logger.error(f"#### Failed to read reservation {settings.RESERVATION_IDENTIFIER} ####")
        return False

    logger.debug(f"#### Successfully read reservation {settings.RESERVATION_IDENTIFIER} ####")
    return reservation_output


async def scontrol_update_reservation(licenses: str, duration: str) -> bool:
    """
    Update reservation duration and licenses.

    Duration format: [days-]hours:minutes:seconds
    Licenses format: foo.bar@server:123,foo.baz@server:234

    Returns True if the reservation was updated successfully, otherwise returns False.
    """
    cmd = [
        str(settings.SCONTROL_PATH),
        "update",
        "reservation",
        f"ReservationName={settings.RESERVATION_IDENTIFIER}",
        f"duration={duration}",
        f"licenses={licenses}",
    ]

    logger.debug(f"#### Updating reservation {settings.RESERVATION_IDENTIFIER} ####")
    try:
        await run_command(cmd)
    except CommandFailedToExecute:
        logger.error(f"#### Failed to update reservation {settings.RESERVATION_IDENTIFIER} ####")
        return False

    logger.debug(f"#### Successfully updated reservation {settings.RESERVATION_IDENTIFIER} ####")
    return True


async def scontrol_delete_reservation() -> bool:
    """
    Delete reservation.

    Returns True if the reservation was deleted successfully, otherwise returns False.
    """
    cmd = [
        str(settings.SCONTROL_PATH),
        "delete",
        f"ReservationName={settings.RESERVATION_IDENTIFIER}",
    ]

    logger.debug(f"#### Deleting reservation {settings.RESERVATION_IDENTIFIER} ####")
    try:
        await run_command(cmd)
    except CommandFailedToExecute:
        logger.error(f"#### Failed to delete reservation {settings.RESERVATION_IDENTIFIER} ####")
        return False

    logger.debug(f"#### Successfully deleted reservation {settings.RESERVATION_IDENTIFIER} ####")
    return True


async def create_or_update_reservation(reservation_data: str):
    """
    Create the reservation if it doesn't exist, otherwise update it.
    If the reservation cannot be updated, delete it and create a new one.
    """
    reservation = await scontrol_show_reservation()

    if reservation:
        updated = await scontrol_update_reservation(reservation_data, "30:00")
        if not updated:
            deleted = await scontrol_delete_reservation()
            LicenseManagerReservationFailure.require_condition(
                deleted, "Could not update or delete reservation."
            )
    else:
        created = await scontrol_create_reservation(reservation_data, "30:00")
        LicenseManagerReservationFailure.require_condition(created, "Could not create reservation.")



"""
Service to clean jobs and bookings that are no longer needed.
"""
import asyncio
from collections import defaultdict
from typing import List, Dict, Tuple

from lm_agent.models import JobSchema, BookingSchema, ConfigurationSchema, LicenseReportItem
from lm_agent.backend_utils.utils import (
    remove_booking,
    remove_job_by_slurm_job_id,
)
from lm_agent.logs import logger
from lm_agent.models import ExtractedBookingSchema, ExtractedUsageSchema


def get_cluster_grace_times(cluster_configurations: List[ConfigurationSchema]) -> Dict[int, int]:
    """
    Get the grace time for each feature_id in the cluster.
    """
    grace_times = {
        feature.id: configuration.grace_time
        for configuration in cluster_configurations
        for feature in configuration.features
    }

    return grace_times


def get_greatest_grace_time_for_job(grace_times: Dict[int, int], job_bookings: List[BookingSchema]) -> int:
    """
    Find the greatest grace time among the features booked by the given job_id.

    If the job has more than one booking, only the greatest grace time will be returned.
    """
    grace_times_for_bookings = (grace_times[booking.feature_id] for booking in job_bookings)

    greatest_grace_time = max(grace_times_for_bookings)

    return greatest_grace_time


def extract_bookings_from_job(job: JobSchema) -> List[ExtractedBookingSchema]:
    """
    Extract all the bookings information from a job.
    """
    return [
        ExtractedBookingSchema(
            booking_id=booking.id,
            job_id=job.id,
            slurm_job_id=job.slurm_job_id,
            username=job.username,
            lead_host=job.lead_host,
            feature_id=booking.feature_id,
            quantity=booking.quantity,
        )
        for booking in job.bookings
    ]


def extract_usages_from_report(report_item: LicenseReportItem) -> List[ExtractedUsageSchema]:
    """
    Extract all the the usage information from a feature report

    Note that the lead_host from the license server comes with the full domain,
    but the lead_host from the job comes without the domain. This is why the
    lead_host is split by the dot and only the first part is used.
    """
    return [
        ExtractedUsageSchema(
            feature_id=report_item.feature_id,
            username=usage.username,
            lead_host=usage.lead_host.split(".")[0],
            quantity=usage.booked,
        )
        for usage in report_item.uses
    ]


def get_bookings_mapping(
    cluster_jobs: List[JobSchema],
) -> Dict[Tuple[int, str, str, int], List[ExtractedBookingSchema]]:
    """
    Map the bookings by creating a key with the required information for the matching.
    """
    bookings_mapping = defaultdict(list)

    for job in cluster_jobs:
        extracted_bookings = extract_bookings_from_job(job)

        for booking in extracted_bookings:
            key = (booking.feature_id, booking.username, booking.lead_host, booking.quantity)
            bookings_mapping[key].append(booking)

    return bookings_mapping


def get_usages_mapping(
    license_report: List[LicenseReportItem],
) -> Dict[Tuple[int, str, str, int], List[ExtractedUsageSchema]]:
    """
    Map the usages by creating a key with the required information for the matching.
    """
    usages_mapping = defaultdict(list)

    for report_item in license_report:
        extracted_usages = extract_usages_from_report(report_item)

        for usage in extracted_usages:
            key = (usage.feature_id, usage.username, usage.lead_host, usage.quantity)
            usages_mapping[key].append(usage)

    return usages_mapping


async def clean_jobs_without_bookings(cluster_jobs: List[JobSchema]) -> List[JobSchema]:
    """
    Clean the jobs that don't have any bookings.
    """
    logger.debug("##### Cleaning jobs without bookings")

    jobs_to_delete = []

    for job in cluster_jobs[:]:
        if not job.bookings:
            jobs_to_delete.append(job.slurm_job_id)
            cluster_jobs.remove(job)

    if not jobs_to_delete:
        logger.debug("##### No jobs without bookings to clean")
        return cluster_jobs

    await asyncio.gather(*[remove_job_by_slurm_job_id(job_id) for job_id in jobs_to_delete])

    logger.debug(f"##### Jobs cleaned: {jobs_to_delete}")
    logger.debug("##### Cleaned jobs without bookings")
    return cluster_jobs


async def clean_jobs_no_longer_running(
    cluster_jobs: List[JobSchema], squeue_result: List[Dict]
) -> List[JobSchema]:
    """
    Clean the jobs that aren't running along with its bookings.
    """
    logger.debug("##### Cleaning jobs that are no longer running")

    jobs_not_running = [str(job["job_id"]) for job in squeue_result if job["state"] != "RUNNING"]
    all_jobs_squeue = [str(job["job_id"]) for job in squeue_result]

    jobs_to_delete = []

    for job in cluster_jobs[:]:
        slurm_job_id = job.slurm_job_id
        if slurm_job_id in jobs_not_running or slurm_job_id not in all_jobs_squeue:
            jobs_to_delete.append(slurm_job_id)
            cluster_jobs.remove(job)

    if not jobs_to_delete:
        logger.debug("##### No need to clean jobs that are no longer running")
        return cluster_jobs

    await asyncio.gather(*[remove_job_by_slurm_job_id(job_id) for job_id in jobs_to_delete])

    logger.debug(f"##### Jobs cleaned: {jobs_to_delete}")
    logger.debug("##### Cleaned jobs that are no longer running")
    return cluster_jobs


async def clean_jobs_by_grace_time(
    cluster_jobs: List[JobSchema], squeue_result: List[Dict], grace_times: Dict[int, int]
) -> List[JobSchema]:
    """
    Clean the jobs where running time is greater than the grace_time.

    If the job has more than one booking, it will be deleted once the
    running time is greater than the greatest grace_time for any of the bookings.
    """
    logger.debug("##### Cleaning jobs by grace time")

    running_jobs = {str(job["job_id"]): job for job in squeue_result if job["state"] == "RUNNING"}

    jobs_to_delete = []

    for job in cluster_jobs[:]:
        greatest_grace_time = get_greatest_grace_time_for_job(grace_times, job.bookings)

        slurm_job_id = job.slurm_job_id

        running_job = running_jobs.get(slurm_job_id)
        if running_job and running_job["run_time_in_seconds"] > greatest_grace_time:
            jobs_to_delete.append(slurm_job_id)
            cluster_jobs.remove(job)

    if not jobs_to_delete:
        logger.debug("##### No jobs to clean by grace time")
        return cluster_jobs

    await asyncio.gather(*[remove_job_by_slurm_job_id(job_id) for job_id in jobs_to_delete])

    logger.debug(f"##### Jobs cleaned: {jobs_to_delete}")
    logger.debug("##### Cleaned jobs by grace time")
    return cluster_jobs


async def clean_bookings_by_usage(cluster_jobs: List[JobSchema], license_report: List[LicenseReportItem]):
    """
    Delete bookings if they match with a usage line in the license report.

    The bookings and the usage lines are mapped by a unique key composed by:
    * feature_id
    * username
    * lead_host
    * quantity

    This will group all bookings and all usages with the same information.

    The bookings can be deleted if the number of usages with the key matches
    the number of bookings with the key.

    - If there are more than one booking matching with the same usage,
    there's no way of knowning which booking relates to the usage.
    In this case, the bookings should be deleted by the grace time clean up.

    - If there are more than one usage matching the same booking,
    the booking could relate to any of the usages.
    In this case, the booking should be deleted by the grace time clean up.

    If there's an equal amount of usages and bookings, all the bookings have
    checked out their licenses from the license server, which means the bookings
    can be safely deleted by this clean up.
    """
    logger.debug("##### Cleaning bookings by usage")

    bookings_mapping = get_bookings_mapping(cluster_jobs)
    usages_mapping = get_usages_mapping(license_report)

    bookings_to_delete = []

    for key, bookings in bookings_mapping.items():
        if len(usages_mapping.get(key, [])) == len(bookings):
            bookings_to_delete.extend([booking.booking_id for booking in bookings])

    if not bookings_to_delete:
        logger.debug("##### No bookings to clean by matching")
        return

    await asyncio.gather(*[remove_booking(booking_id) for booking_id in bookings_to_delete])
    logger.debug(f"##### Bookings cleaned: {bookings_to_delete}")
    logger.debug("##### Cleaned bookings by matching")


async def clean_jobs_and_bookings(
    cluster_configurations: List[ConfigurationSchema],
    cluster_jobs: List[JobSchema],
    squeue_result: List[Dict],
    license_report: List[LicenseReportItem],
):
    """
    Clean the jobs and bookings that are no longer needed.

    The jobs can be deleted by:
    * Deleting the jobs that don't have any bookings.
    * Deleting the jobs that are no longer running.
    * Deleting the jobs that are running longer than the grace time.

    The bookings can be deleted by:
    * Deleting the bookings that have checked out their licenses from the license server.
    """
    logger.debug("##### Start cleaning jobs and bookings")

    grace_times = get_cluster_grace_times(cluster_configurations)

    jobs_with_bookings = await clean_jobs_without_bookings(cluster_jobs)
    jobs_still_running = await clean_jobs_no_longer_running(jobs_with_bookings, squeue_result)
    jobs_within_grace_time = await clean_jobs_by_grace_time(jobs_still_running, squeue_result, grace_times)
    await clean_bookings_by_usage(jobs_within_grace_time, license_report)

    logger.debug("##### Finished cleaning jobs and bookings")



"""Utilities that interact with slurm."""
import re
from typing import Dict, List, Optional, Union

from lm_agent.exceptions import ScontrolRetrievalFailure, SqueueParserUnexpectedInputError
from lm_agent.models import LicenseBooking
from lm_agent.logs import logger, log_error
from lm_agent.config import settings
from lm_agent.utils import run_command


def _match_requested_license(requested_license: str) -> Union[dict, None]:
    license_regex = re.compile(
        r"(?P<product>[a-zA-Z0-9-_]+)\.(?P<feature>[a-zA-Z0-9-_]+)@(?P<server_type>\w+)(:(?P<tokens>\d+))?"
    )

    matches = license_regex.match(requested_license)

    if not matches:
        return None

    groups = matches.groupdict()
    if not groups["tokens"]:
        groups["tokens"] = "1"

    return {
        "product_feature": groups["product"] + "." + groups["feature"],
        "server_type": groups["server_type"],
        "tokens": int(groups["tokens"]),
    }


def get_required_licenses_for_job(job_licenses: str) -> List:
    """Retrieve the required licenses for a job."""

    license_array = job_licenses.split(",")
    logger.debug(f"##### License array for job: {license_array} #####")

    required_licenses: List = []

    if not license_array:
        return required_licenses

    if license_array[0] == "(null)":
        return required_licenses

    for requested_license in license_array:
        matched_license_items = _match_requested_license(requested_license)
        if not matched_license_items:
            continue
        product_feature = matched_license_items["product_feature"]
        booked = matched_license_items["tokens"]

        # Create the license booking
        license_booking = LicenseBooking(
            product_feature=product_feature,
            quantity=booked,
        )
        required_licenses.append(license_booking)

    return required_licenses


async def get_all_features_cluster_values() -> Optional[Dict[str, Dict[str, int]]]:
    """
    Parse the output from `scontrol show lic` and return a dictionary of
    product_feature: {"total": <total>, "used": <used>}.

    Note: the line with the counters doesn't include the feature name,
    so to find the feature it's necessary to look at the previous line.
    """
    product_feature_line = re.compile(
        r"LicenseName=(?P<product>[a-zA-Z0-9-_]+).(?P<feature>[a-zA-Z0-9-_]+)@(?P<license_server_type>\w+)"
    )

    used_tokens_line = re.compile(
        r"^\s*Total=(?P<total>\d+) Used=(?P<used>\d+) Free=(?P<free>\d+) Reserved=(?P<reserved>\d+) Remote=(?P<remote>\w+)"  # noqa
    )

    scontrol_output = await scontrol_show_lic()

    parsed_data: dict = {}
    product_feature_list: list = []

    for line in scontrol_output.split("\n"):
        parsed_product_feature = product_feature_line.match(line)
        if parsed_product_feature:
            product_feature_data = parsed_product_feature.groupdict()
            product_feature = product_feature_data["product"] + "." + product_feature_data["feature"]
            product_feature_list.append(product_feature)
            continue

        parsed_used_tokens = used_tokens_line.match(line)
        if parsed_used_tokens:
            used_data = parsed_used_tokens.groupdict()
            product_feature = product_feature_list[-1]
            parsed_data[product_feature] = {"total": int(used_data["total"]), "used": int(used_data["used"])}
            continue

    return parsed_data


async def scontrol_show_lic():
    """
    Get the license usage from scontrol.
    """

    cmd = [
        str(settings.SCONTROL_PATH),
        "show",
        "lic",
    ]
    output = await run_command(cmd)

    logger.debug("##### scontrol show lic #####")
    return output


async def get_lead_host(nodelist):
    """
    Get the job's lead host from the nodelist.

    The lead host is the first node in the nodelist.
    The nodelist can contain multiple lists of nodes inside square brackets.
    """
    cmd = [
        str(settings.SCONTROL_PATH),
        "show",
        "hostnames",
        nodelist,
    ]
    output = await run_command(cmd)

    lead_host = output.split("\n")[0]
    ScontrolRetrievalFailure.require_condition(lead_host != "", "Could not get lead host from nodelist")

    return lead_host


async def get_all_product_features_from_cluster() -> List[str]:
    """
    Returns a list of all product.feature in the cluster.
    """
    show_lic_output = await scontrol_show_lic()

    PRODUCT_FEATURE = r"LicenseName=(?P<product>[a-zA-Z0-9-_]+).(?P<feature>[a-zA-Z0-9-_]+)"
    RX_PRODUCT_FEATURE = re.compile(PRODUCT_FEATURE)

    parsed_features = []
    output = show_lic_output.split("\n")
    for line in output:
        parsed_line = RX_PRODUCT_FEATURE.match(line)
        if parsed_line:
            parsed_data = parsed_line.groupdict()
            product = parsed_data["product"]
            feature = parsed_data["feature"]
            parsed_features.append(f"{product}.{feature}")

    return parsed_features


async def return_formatted_squeue_out() -> str:
    """
    Call squeue via Popen and return the formatted output.

    Return the squeue output in the form "<job_id>|<run_time>|<state>".
    """
    cmd = [
        str(settings.SQUEUE_PATH),
        "--noheader",
        "--format='%A|%M|%T'",
    ]

    return await run_command(cmd)


def _total_time_in_seconds(time_string: str) -> int:
    """
    Return the runtime in seconds for a job.

    This function takes a slurm time string ("<days>-<hours>:<minutes>:<seconds>") as input, parses
    and converts each of the units in the time string to seconds and returns a computed value, the sum of
    the days, hours, minutes and seconds (in seconds).
    """
    MINUTE = 60
    HOUR = 60 * MINUTE
    DAY = 24 * HOUR

    days = 0
    hours = 0
    splitted_time = [int(value) for value in re.split("-|:", time_string)]
    splitted_time_len = len(splitted_time)

    if splitted_time_len == 4:
        days, hours, minutes, seconds = splitted_time
    elif splitted_time_len == 3:
        hours, minutes, seconds = splitted_time
    else:
        minutes, seconds = splitted_time

    return days * DAY + hours * HOUR + minutes * MINUTE + seconds


def squeue_parser(squeue_formatted_output) -> List:
    """Parse the squeue formatted output."""

    squeue_parsed_output: List = []

    if not squeue_formatted_output:
        return squeue_parsed_output

    def parse_squeue_line():
        """Parse a line from squeue formatted output."""
        with SqueueParserUnexpectedInputError.handle_errors(
            "Unexpected input from squeue", do_except=log_error
        ):
            job_id, run_time, state = line.strip("'").split("|")

        return job_id, run_time, state

    for line in squeue_formatted_output.split():
        job_id, run_time, state = parse_squeue_line()
        squeue_parsed_output.append(
            {
                "job_id": int(job_id),
                "run_time_in_seconds": _total_time_in_seconds(run_time),
                "state": state,
            }
        )

    return squeue_parsed_output



"""
Parser for LM-X
"""
import re
from typing import Optional
from lm_agent.models import LicenseUsesItem


HOSTWORD = r"[a-zA-Z0-9-]+"
HOSTNAME = rf"{HOSTWORD}(\.{HOSTWORD})*"
INT = r"\d+"
VERSION = rf"{INT}\.{INT}"

FEATURE_LINE = rf"^Feature: (?P<feature>\w+) Version: (?P<version>{VERSION}) Vendor: (?P<vendor>\S+)$"
IN_USE_LINE = rf"^(?P<in_use>{INT}) of (?P<total>{INT}) license\(s\) used(:)?$"
USAGE_LINE_1 = (
    rf"^(?P<in_use>{INT}) license\(s\) "
    rf"used by (?P<user>\S+)@(?P<lead_host>{HOSTNAME}) "
    rf"\[{INT}\.{INT}\.{INT}\.{INT}\]"
)
USAGE_LINE_2 = (
    rf"^(?P<in_use>{INT}) license\(s\) "
    rf"used by (?P<user>\S+)@(?P<lead_host>{HOSTNAME})_\S+ "
    rf"\[{INT}\.{INT}\.{INT}\.{INT}\]"
)
RX_FEATURE = re.compile(FEATURE_LINE)
RX_IN_USE = re.compile(IN_USE_LINE)
RX_USAGE_1 = re.compile(USAGE_LINE_1)
RX_USAGE_2 = re.compile(USAGE_LINE_2)
USAGE_LINES = [RX_USAGE_1, RX_USAGE_2]


def parse_feature_line(line: str) -> Optional[str]:
    """
    Parse the feature line in the LM-X output.
    Data we need:
    - ``feature``: license name
    """
    parsed_feature = RX_FEATURE.match(line)
    if parsed_feature is None:
        return None
    feature_data = parsed_feature.groupdict()

    return feature_data["feature"].lower()


def parse_in_use_line(line: str) -> Optional[dict]:
    """
    Parse the in use line in the LM-X output.
    Data we need:
    - ``in_use``: quantity of licenses being use
    - ``total``: total amount of licenses

    Obs: this line doesn't include the license name.
    The license in use is the last one parsed before this line.
    """

    parsed_in_use = RX_IN_USE.match(line)
    if parsed_in_use is None:
        return None
    in_use_data = parsed_in_use.groupdict()

    return {
        "used": int(in_use_data["in_use"]),
        "total": int(in_use_data["total"]),
    }


def parse_usage_line(line: str) -> Optional[LicenseUsesItem]:
    """
    Parse the usage line in the LS-Dyna output.
    Data we need:
    -``username``: user who booked the license
    -``lead_host``: host using the license
    -``booked``: quantity of licenses being used

    Obs: this line also doesn't incluse the license name.
    The license in use is the last one parsed before this line.
    """
    for RX in USAGE_LINES:
        parsed_usage = RX.match(line)
        if parsed_usage:
            break

    if parsed_usage is None:
        return None
    usage_data = parsed_usage.groupdict()

    return LicenseUsesItem(
        username=usage_data["user"].lower(),
        lead_host=usage_data["lead_host"],
        booked=int(usage_data["in_use"]),
    )


def parse(server_output: str) -> dict:
    """
    Parse the LM-X output using regex to match the lines we need:
    -``feature line``: info about each license
    -``in use line``: info about licenses in use
    -``usage line``: info about users using licenses

    Since the in use and usage line don't have the name of the license in use,
    we're saving each parsed license in a list. This way, we can find which
    license is being used by checking the last parsed license in the list.
    """
    parsed_data: dict = {}
    feature_list: list = []

    for line in server_output.splitlines():
        parsed_feature = parse_feature_line(line)
        parsed_in_use = parse_in_use_line(line)
        parsed_usage = parse_usage_line(line)

        if parsed_feature:
            parsed_data[parsed_feature] = {}
            feature_list.append(parsed_feature)

        elif parsed_in_use:
            feature = feature_list[-1]
            parsed_data[feature] = {
                "used": parsed_in_use["used"],
                "total": parsed_in_use["total"],
                "uses": [],
            }
        elif parsed_usage:
            feature = feature_list[-1]
            parsed_data[feature]["uses"].append(parsed_usage)
        else:
            continue

    return parsed_data
